# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Decorators for testing certain assertions."""

from __future__ import absolute_import, print_function

from functools import wraps

from flask import request

from .errors import InvalidContentType


def require_content_types(*allowed_content_types):
    r"""Decorator to test if proper Content-Type is provided.

    :param \*allowed_content_types: List of allowed content types.
    :raises invenio_rest.errors.InvalidContentType: It's rised if a content
        type not allowed is required.
    """
    def decorator(f):
        @wraps(f)
        def inner(*args, **kwargs):
            if request.mimetype not in allowed_content_types:
                raise InvalidContentType(allowed_content_types)
            return f(*args, **kwargs)
        return inner
    return decorator



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from __future__ import absolute_import, print_function

import json

from invenio_rest import InvenioREST
from invenio_rest.decorators import require_content_types


def test_require_content_types(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route("/", methods=['POST'])
    @require_content_types('application/json', 'text/plain')
    def test_view():
        return "OK"

    with app.test_client() as client:
        res = client.post("/", content_type='application/json', data="{}")
        assert res.status_code == 200
        res = client.post("/", content_type='text/plain', data="test")
        assert res.status_code == 200
        res = client.post("/", content_type='application/json;charset=utf-8',
                          data="{}")
        assert res.status_code == 200
        res = client.post("/", content_type='application/xml', data="<d></d>")
        assert res.status_code == 415
        data = json.loads(res.get_data(as_text=True))
        assert data['status'] == 415
        assert 'application/json' in data['message']
        assert 'text/plain' in data['message']



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import json
from datetime import datetime

import pkg_resources
import pytest
from flask import Flask, abort, make_response, request
from flask.json import jsonify
from six.moves.urllib.parse import urlencode
from werkzeug.http import quote_etag, unquote_etag

from invenio_rest import ContentNegotiatedMethodView, InvenioREST
from invenio_rest.errors import RESTException


def test_version():
    """Test version import."""
    from invenio_rest import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = InvenioREST(app)
    assert 'invenio-rest' in app.extensions

    app = Flask('testapp')
    ext = InvenioREST()
    assert 'invenio-rest' not in app.extensions
    ext.init_app(app)
    assert 'invenio-rest' in app.extensions


def test_error_handlers(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route("/<int:status_code>",
               methods=['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD',
                        'TRACE', 'OPTIONS'])
    def test_view(status_code):
        abort(status_code)

    error_codes = [
        400, 401, 403, 404, 405, 406, 409, 410, 412, 422, 415, 429, 500, 501,
        502, 503, 504]

    with app.test_client() as client:
        verbs = [client.get, client.post, client.put, client.delete,
                 client.patch, client.trace, client.options, client.get]
        for s in error_codes:
            for verb in verbs:
                res = verb("/{0}".format(s))
                assert res.status_code == s
                if pkg_resources.get_distribution("werkzeug").\
                        version == '0.14.1' and s == 412:
                    # skip test because of the following issue:
                    # https://github.com/pallets/werkzeug/issues/1231
                    continue
                if verb != client.head:
                    data = json.loads(res.get_data(as_text=True))
                    assert data['status'] == s
                    assert data['message']


def test_custom_httpexception(app):
    """Test custom RESTException."""
    InvenioREST(app)

    class CustomBadRequest(RESTException):
        code = 400
        description = 'test'

    @app.route('/')
    def test_view():
        raise CustomBadRequest()

    with app.test_client() as client:
        res = client.get('/')
        assert res.status_code == 400
        data = json.loads(res.get_data(as_text=True))
        assert data['message'] == 'test'
        assert data['status'] == 400


def test_cors(app):
    """Test CORS support."""
    app.config['REST_ENABLE_CORS'] = True
    app.config['CORS_ALWAYS_SEND'] = False

    InvenioREST(app)

    @app.route('/')
    def cors_test():
        return 'test'

    with app.test_client() as client:
        res = client.get('/')
        assert not res.headers.get('Access-Control-Allow-Origin', False)
        assert not res.headers.get('Access-Control-Expose-Headers', False)
        res = client.get('/', headers=[('Origin', 'http://example.com')])
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert len(res.headers['Access-Control-Expose-Headers'].split(',')) \
            == len(app.config['CORS_EXPOSE_HEADERS'])


def _obj_to_json_serializer(data, code=200, headers=None):
    if data:
        res = jsonify(data)
        res.cache_control.no_cache = True
        res.set_etag('abc')
    else:
        res = make_response()
    res.status_code = code
    return res


class _ObjectItem(ContentNegotiatedMethodView):
    view_name = 'object_item'

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(_ObjectItem, self).__init__(*args, **kwargs)
        self.modified = datetime.utcnow()

    def get(self, id, **kwargs):
        """GET a resource."""
        self.check_etag('abc', weak='weak_etags' in request.args)
        self.check_if_modified_since(self.modified)
        if id == 42:
            abort(404)
        # return {'id': id, 'method': 'GET'}, 200
        res = self.make_response({'id': id, 'method': 'GET'}, 200)
        res.last_modified = self.modified
        return res

    def put(self, id, **kwargs):
        """PUT a resource."""
        self.check_etag('abc', weak='weak_etags' in request.args)
        if id == 42:
            return self.make_response(None, code=404)
        return self.make_response({'id': id, 'method': 'PUT'}, 200)

    def patch(self, id, **kwargs):
        """PATCH a resource."""
        self.check_etag('abc', weak='weak_etags' in request.args)
        if id == 42:
            return None, 404
        return self.make_response({'id': id, 'method': 'PATCH'})

    def post(self, id, **kwargs):
        """POST a resource."""
        self.check_etag('abc', weak='weak_etags' in request.args)
        if id == 42:
            abort(404)
        return {'id': id, 'method': 'POST'}

    def delete(self, id, **kwargs):
        """DELETE a resource."""
        self.check_etag('abc', weak='weak_etags' in request.args)
        if id == 42:
            abort(404)
        return {'id': id, 'method': 'DELETE'}


def test_content_negotiation_method_view_global_serializers(app):
    """Test ContentNegotiationMethodView. Test default serializers."""
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'serializers': {
            'application/json': _obj_to_json_serializer,
        }
    })


def test_content_negotiation_method_view_method_serializers(app):
    """Test method serializers with no global serializers."""
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'method_serializers': {
            'GET': {'application/json': _obj_to_json_serializer, },
            'PUT': {'application/json': _obj_to_json_serializer, },
            'POST': {'application/json': _obj_to_json_serializer, },
            'PATCH': {'application/json': _obj_to_json_serializer, },
            'DELETE': {'application/json': _obj_to_json_serializer, },
        }
    })


def test_content_negotiation_method_view_method_serializers_2(app):
    """Check that method serializers override global ones."""
    def failing_serializer(data, code=200, headers=None):
        raise Exception('Method serializer was not called.')

    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'serializers': {
            'application/json': failing_serializer,
        },
        'method_serializers': {
            'GET': {'application/json': _obj_to_json_serializer, },
            'PUT': {'application/json': _obj_to_json_serializer, },
            'POST': {'application/json': _obj_to_json_serializer, },
            'PATCH': {'application/json': _obj_to_json_serializer, },
            'DELETE': {'application/json': _obj_to_json_serializer, },
        }
    })


def test_content_negotiation_method_view_mixed_serializers(app):
    """Test ContentNegotiationMethodView. Test partial method serializers"""
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'serializers': {
            'application/json': _obj_to_json_serializer,
        },
        'method_serializers': {
            'GET': {'application/json': _obj_to_json_serializer, },
            'PUT': {'application/json': _obj_to_json_serializer, },
        }
    })


def test_default_media_type(app):
    """Test setting default media type."""
    def failing_serializer(data, code=200, headers=None):
        raise Exception('Wrong serializer was called.')
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'serializers': {
            'application/mpeg': failing_serializer,
            'application/json': _obj_to_json_serializer,
        },
        'default_media_type': 'application/json'
    })


def test_default_media_type_for_all_methods(app):
    """Test setting default serializer for all methods."""
    def failing_serializer(data, code=200, headers=None):
        raise Exception('Wrong method serializer was called.')
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'method_serializers': {
            'GET': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'PUT': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'POST': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'PATCH': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'DELETE': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
        },
        'default_media_type': 'application/json'
    })


def test_default_method_media_type(app):
    """Test setting default media type for each method."""
    def failing_serializer(data, code=200, headers=None):
        raise Exception('Wrong method serializer was called.')
    _subtest_content_negotiation_method_view(app, _ObjectItem, params={
        'method_serializers': {
            'GET': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'PUT': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'POST': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'PATCH': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
            'DELETE': {
                'application/json': _obj_to_json_serializer,
                'application/mpeg': failing_serializer,
            },
        },
        'default_method_media_type': {
            'GET': 'application/json',
            'PUT': 'application/json',
            'POST': 'application/json',
            'PATCH': 'application/json',
            'DELETE': 'application/json',
        },
        'default_media_type': 'application/doesnotexist'
    })


def test_default_method_media_type_error(app):
    """Test that there is an error when multiple serializers are given with no
    default media type.
    """
    with pytest.raises(ValueError):
        _ObjectItem(
            method_serializers={
                'GET': {
                    'application/json': _obj_to_json_serializer,
                    'application/json-patch+json': _obj_to_json_serializer,
                },
                'PUT': {
                    'application/json': _obj_to_json_serializer,
                    'application/json-patch+json': _obj_to_json_serializer,
                },
                'POST': {
                    'application/json': _obj_to_json_serializer,
                    'application/json-patch+json': _obj_to_json_serializer,
                },
                'PATCH': {
                    'application/json': _obj_to_json_serializer,
                    'application/json-patch+json': _obj_to_json_serializer,
                },
                'DELETE': {
                    'application/json': _obj_to_json_serializer,
                    'application/json-patch+json': _obj_to_json_serializer,
                },
            },
        )


def test_default_media_type_error(app):
    """Test that there is an error when multiple serializers are given with no
    default media type.
    """
    with pytest.raises(ValueError):
        _ObjectItem(serializers={
            'application/json': _obj_to_json_serializer,
            'application/json-patch+json': _obj_to_json_serializer,
        })


def test_get_method_serializers(app):
    """Test get request method serializers."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            DELETE={'*/*': 'any-delete', },
            GET={'text/plain': 'plain-get', 'text/html': 'html-get'},
            POST={'text/plain': 'plain-post', 'text/html': 'html-post'}
        ),
        serializers={
            'text/plain': 'plain',
            'text/html': 'html',
        },
        default_media_type='application/json',
        default_method_media_type=dict(
            GET='text/plain',
            POST='text/html',
        )
    )
    # Default serializers
    assert v.get_method_serializers('GET')[0] == v.method_serializers['GET']
    assert v.get_method_serializers('HEAD')[0] == v.method_serializers['GET']
    assert v.get_method_serializers('POST')[0] == v.method_serializers['POST']
    assert v.get_method_serializers('PUT')[0] == v.serializers

    # Default media type
    assert v.get_method_serializers('GET')[1] == \
        v.default_method_media_type['GET']
    assert v.get_method_serializers('HEAD')[1] == \
        v.default_method_media_type['GET']
    assert v.get_method_serializers('POST')[1] == \
        v.default_method_media_type['POST']
    assert v.get_method_serializers('PUT')[1] == v.default_media_type

    assert v.get_method_serializers('DELETE')[0] == \
        v.method_serializers['DELETE']


def test_match_serializers_headers(app):
    """Test match serializers headers."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            DELETE={
                '*/*': 'any-delete',
            },
            GET={
                'application/json': 'json-get',
                'application/marcxml+xml': 'xml-get'},
            POST={
                'application/json': 'json-post',
                'application/marcxml+xml': 'xml-post'}
        ),
        serializers={
            'application/json': 'json',
            'application/marcxml+xml': 'xml',
        },
        default_media_type='application/json',
        default_method_media_type=dict(
            GET='application/json',
            POST='application/marcxml+xml',
        )
    )

    tests = [
        # Should choose application/json
        ('text/plain,application/json,*/*', 'GET', 'json-get'),
        ('text/plain,application/json,*/*', 'HEAD', 'json-get'),
        ('text/plain,application/json,*/*', 'POST', 'json-post'),
        ('text/plain,application/json,*/*', 'PUT', 'json'),
        # Should choose application/json (even with lower quality)
        ('text/plain,application/json; q=0.5, */*', 'GET', 'json-get'),
        ('text/plain,application/json; q=0.5, */*', 'HEAD', 'json-get'),
        ('text/plain,application/json; q=0.5, */*', 'POST', 'json-post'),
        ('text/plain,application/json; q=0.5, */*', 'PUT', 'json'),
        # Should choose application/marcxml+xml
        ('text/plain,application/marcxml+xml,*/*', 'GET', 'xml-get'),
        ('text/plain,application/marcxml+xml,*/*', 'HEAD', 'xml-get'),
        ('text/plain,application/marcxml+xml,*/*', 'POST', 'xml-post'),
        ('text/plain,application/marcxml+xml,*/*', 'PUT', 'xml'),
        # Should choose default defined by default_media_type and
        # default_method_media_type
        ('text/plain,*/*', 'GET', 'json-get'),
        ('text/plain,*/*', 'HEAD', 'json-get'),
        ('text/plain,*/*', 'POST', 'xml-post'),
        ('text/plain,*/*', 'PUT', 'json'),
        # Should choose none
        ('text/plain', 'GET', None),
        ('text/plain', 'HEAD', None),
        ('text/plain', 'POST', None),
        ('text/plain', 'PUT', None),
        # Should choose default
        (None, 'GET', 'json-get'),
        (None, 'HEAD', 'json-get'),
        (None, 'POST', 'xml-post'),
        (None, 'PUT', 'json'),
        # Should choose application/json
        ('application/json', 'GET', 'json-get'),
        # Should choose highest quality
        ('application/json; q=0.5, application/marcxml+xml', 'GET', 'xml-get'),
        ('application/json, application/marcxml+xml; q=0.5', 'GET',
         'json-get'),
        ('application/json; q=0.4, application/marcxml+xml; q=0.6', 'GET',
         'xml-get'),
        ('application/marcxml+xml; q=0.6, application/json; q=0.4', 'GET',
         'xml-get'),
        # Test delete
        ('application/marcxml+xml; q=0.6, application/json; q=0.4', 'DELETE',
         'any-delete'),
        ('video/mp4', 'DELETE', 'any-delete'),
    ]

    for accept, method, expected_serializer in tests:
        params = dict(headers=[('Accept', accept)]) if accept else {}
        _test_march_serializers(app, v, params, method, accept,
                                expected_serializer)


def test_match_serializers_query_arg(app):
    """Test match serializers query argument."""
    v = ContentNegotiatedMethodView(
        method_serializers=dict(
            GET={
                'application/json': 'json-get',
                'application/marcxml+xml': 'xml-get'}
        ),
        serializers_query_aliases={
            'json': 'application/json',
            'marcxml': 'application/marcxml+xml',
        },
        default_media_type='application/marcxml+xml',
        default_method_media_type=dict(
            GET='application/marcxml+xml',
        )
    )

    # enable serialization by query arg
    app.config['REST_MIMETYPE_QUERY_ARG_NAME'] = 'format'
    InvenioREST(app)

    arg_name = app.config['REST_MIMETYPE_QUERY_ARG_NAME']
    tests = [
        # Should serialize to marcxml
        (urlencode({arg_name: 'marcxml'}), 'application/json', 'xml-get'),
        # Should serialize to json
        (urlencode({arg_name: 'json'}), 'application/marcxml+xml',
         'json-get'),
        # Should serialize to json
        (urlencode({}), 'application/json', 'json-get'),
        # Should serialize to default (marcxml)
        (urlencode({'wrong_name': 'marcxml'}), None, 'xml-get'),
    ]

    for arg, accept, expected_serializer in tests:
        params = dict(headers=[('Accept', accept)],
                      query_string=arg) if accept else {}
        _test_march_serializers(app, v, params, 'GET', accept,
                                expected_serializer)

    # disable query arg, should serialize by headers even if query arg passed
    app.config['REST_MIMETYPE_QUERY_ARG_NAME'] = None
    InvenioREST(app)

    accept = 'application/json'
    params = dict(headers=[('Accept', accept)], query_string='format=marcxml')
    _test_march_serializers(app, v, params, 'GET', accept, 'json-get')


def _test_march_serializers(app, view_method, request_params, request_method,
                            request_accept_header, expected_serializer):
    with app.test_request_context(**request_params):
        assert view_method.match_serializers(
            *view_method.get_method_serializers(request_method)) == \
               expected_serializer, "Accept: {0}, Method: {1}".format(
            request_accept_header, request_method)


def _subtest_content_negotiation_method_view(app, content_negotiated_class,
                                             params):
    """Test a given ContentNegotiatedMethodView subclass."""
    app.add_url_rule('/objects/<int:id>',
                     view_func=content_negotiated_class
                     .as_view(content_negotiated_class.view_name,
                              **params))

    with app.test_client() as client:
        read_methods = [client.get, client.head]
        write_methods = [client.patch, client.put, client.post, client.delete]
        all_methods = read_methods + write_methods
        method_names = {
            'GET': client.get,
            'POST': client.post,
            'PUT': client.put,
            'PATCH': client.patch,
            'DELETE': client.delete,
        }

        def check_normal_response(res, method):
            if method != client.head:
                parsed = json.loads(res.get_data(as_text=True))
                expected = {'id': 1, 'method': parsed['method']}
                assert parsed == expected
                # check that the right method was called
                assert method_names[parsed['method']] == method
                assert res.content_type == 'application/json'
            assert res.status_code == 200
            # check that Cache-Control header is set
            assert res.cache_control.no_cache
            # check that the ETag is correct
            assert unquote_etag(res.headers['ETag']) == \
                unquote_etag(quote_etag('abc'))

        def check_304_response(res):
            assert res.status_code == 304
            # check that Cache-Control header is set
            assert res.cache_control.no_cache
            # check that the ETag is correct
            assert unquote_etag(res.headers['ETag']) == \
                unquote_etag(quote_etag('abc'))

        # check valid call with condition
        headers = [('Accept', 'application/json')]
        for method in all_methods:
            res = method('/objects/1', headers=headers)
            check_normal_response(res, method)

        # check valid call without condition
        for method in all_methods:
            res = method('/objects/1')
            check_normal_response(res, method)

        # check that non accepted mime types are not accepted
        headers = [('Accept', 'application/xml')]
        for method in all_methods:
            res = method('/objects/1', headers=headers)
            assert res.status_code == 406

        # check that errors are forwarded properly
        headers = [('Accept', 'application/json')]
        for method in all_methods:
            res = method('/objects/42', headers=headers)
            assert res.status_code == 404

        # check Matching If-None-Match
        headers_nonmatch_match = [('Accept', 'application/json'),
                                  ('If-None-Match', '"xyz", "abc"')]
        headers_nonmatch_star = [('Accept', 'application/json'),
                                 ('If-None-Match', '"xyz", "*"')]
        for method in read_methods:
            res = method('/objects/1', headers=headers_nonmatch_match)
            check_304_response(res)
            res = method('/objects/1', headers=headers_nonmatch_star)
            check_304_response(res)

        for method in write_methods:
            res = method('/objects/1', headers=headers_nonmatch_match)
            assert res.status_code == 412
            res = method('/objects/1', headers=headers_nonmatch_star)
            assert res.status_code == 412

        # check Matching If-None-Match with weak ETags
        headers_nonmatch_match = [('Accept', 'application/json'),
                                  ('If-None-Match', 'W/"xyz", W/"abc"')]
        headers_nonmatch_star = [('Accept', 'application/json'),
                                 ('If-None-Match', 'W/"xyz", "*"')]
        query_string_weak_etags = {'weak_etags': True}
        for method in read_methods:
            res = method('/objects/1',
                         headers=headers_nonmatch_match,
                         query_string=query_string_weak_etags)
            check_304_response(res)
            res = method('/objects/1',
                         headers=headers_nonmatch_star,
                         query_string=query_string_weak_etags)
            check_304_response(res)

        for method in write_methods:
            res = method('/objects/1',
                         headers=headers_nonmatch_match,
                         query_string=query_string_weak_etags)
            assert res.status_code == 412
            res = method('/objects/1',
                         headers=headers_nonmatch_star,
                         query_string=query_string_weak_etags)
            assert res.status_code == 412

        # check non matching If-None-Match
        headers = [('Accept', 'application/json'),
                   ('If-None-Match', '"xyz", "def"')]
        for method in all_methods:
            res = method('/objects/1', headers=headers)
            check_normal_response(res, method)

        # check non matching If-None-Match with weak ETags
        headers = [('Accept', 'application/json'),
                   ('If-None-Match', 'W/"xyz", W/"def"')]
        query_string_weak_etags = {'weak_etags': True}
        for method in all_methods:
            res = method('/objects/1',
                         headers=headers,
                         query_string=query_string_weak_etags)
            check_normal_response(res, method)

        # check matching If-Match
        headers = [('Accept', 'application/json'),
                   ('If-Match', '"abc", "def"')]
        for method in all_methods:
            res = method('/objects/1', headers=headers)
            check_normal_response(res, method)

        # check matching If-Match with weak ETags
        headers = [('Accept', 'application/json'),
                   ('If-Match', 'W/"abc", W/"def"')]
        query_string_weak_etags = {'weak_etags': True}
        for method in all_methods:
            res = method('/objects/1',
                         headers=headers,
                         query_string=query_string_weak_etags)
            check_normal_response(res, method)

        # check non matching If-Match
        headers = [('Accept', 'application/json'),
                   ('If-Match', '"xyz", "def"')]
        for method in all_methods:
            res = method('/objects/1', headers=headers)
            assert res.status_code == 412

        # check non matching If-Match with weak ETags
        headers = [('Accept', 'application/json'),
                   ('If-Match', 'W/"xyz", W/"def"')]
        query_string_weak_etags = {'weak_etags': True}
        for method in all_methods:
            res = method('/objects/1',
                         headers=headers,
                         query_string=query_string_weak_etags)
            assert res.status_code == 412

        # check If-Modified-Since
        res = client.get('/objects/1')
        last_modified = res.headers['Last-Modified']
        res = client.get(
            '/objects/1', headers={'If-Modified-Since': last_modified})
        assert res.cache_control.no_cache
        assert res.status_code == 304



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import shutil
import tempfile

import pytest
from flask import Flask

from invenio_rest.ext import InvenioREST


@pytest.fixture()
def instance_path():
    """Temporary instance path."""
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path)


@pytest.fixture()
def base_app(instance_path):
    """Flask application fixture."""
    app_ = Flask('testapp', instance_path=instance_path)
    app_.config.update(
        SECRET_KEY='SECRET_KEY',
        TESTING=True,
        REST_CSRF_ENABLED=False,
    )

    @app_.route('/ping')
    def ping():
        return 'ping'

    return app_


@pytest.fixture()
def app(base_app):
    """Flask application fixture."""
    with base_app.app_context():
        yield base_app


@pytest.fixture()
def csrf_app(base_app):
    """Flask application fixture with csrf enabled."""
    InvenioREST(base_app)

    @base_app.route('/csrf-protected', methods=["POST"])
    def csrf_test():
        return 'test'

    with base_app.app_context():
        yield base_app


@pytest.fixture()
def csrf(csrf_app):
    """Initialize CSRF object on every test function.

    We don't import `csrf` from `invenio_rest.csrf` because
    like that it is not cleared after every test fuction run.
    """
    from invenio_rest.csrf import CSRFProtectMiddleware
    csrf = CSRFProtectMiddleware(csrf_app)
    assert 'invenio-csrf' in csrf_app.extensions
    yield csrf



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API serializers."""

import warnings

from marshmallow import Schema


class MarshmalDict(dict):
    """Wrapping class for result of type dictionary."""

    def __init__(self, result):
        """Initialize MarshmalDict."""
        self.update(result)

    @property
    def data(self):
        """Substituting data property for backwards compatibility."""
        warnings.warn(
            "Schema().dump().data and Schema().dump().errors "
            "as well as Schema().load().data and Schema().loads().data"
            "attributes are deprecated in marshmallow v3.x.",
            category=PendingDeprecationWarning, stacklevel=2)
        return self


class MarshmalList(list):
    """Wrapping class for result of type list."""

    def __init__(self, result):
        """Initialize MarshmalList."""
        self.extend(result)

    @property
    def data(self):
        """Substituting data property for backwards compatibility."""
        warnings.warn(
            "Schema().dump().data and Schema().dump().errors "
            "as well as Schema().load().data and Schema().loads().data"
            "attributes are deprecated in marshmallow v3.x.",
            category=PendingDeprecationWarning, stacklevel=2)
        return self


def result_wrapper(result):
    """Wrap schema returned dump value."""
    if isinstance(result, tuple):
        return result
    elif isinstance(result, dict):
        return MarshmalDict(result)
    elif isinstance(result, list):
        return MarshmalList(result)
    return result


class BaseSchema(Schema):
    """Base schema for all serializations."""

    def dump(self, obj, *args, **kwargs):
        """Wrap dump result for backward compatibility."""
        result = super(BaseSchema, self).dump(obj, **kwargs)
        return result_wrapper(result)

    def dumps(self, obj, *args, **kwargs):
        """Wrap dumps result for backward compatibility."""
        result = super(BaseSchema, self).dumps(obj, *args, **kwargs)
        return result_wrapper(result)

    def load(self, obj, *args, **kwargs):
        """Wrap load result for backward compatibility."""
        result = super(BaseSchema, self).load(obj, *args, **kwargs)
        return result_wrapper(result)

    def loads(self, obj, *args, **kwargs):
        """Wrap loads result for backward compatibility."""
        result = super(BaseSchema, self).loads(obj, *args, **kwargs)
        return result_wrapper(result)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Error module tests."""

from __future__ import absolute_import, print_function

import json

from invenio_rest import InvenioREST
from invenio_rest.errors import FieldError, InvalidContentType, \
    RESTException, RESTValidationError


def test_errors(app):
    """Error handlers view."""
    InvenioREST(app)

    @app.route('/', methods=['GET'])
    def test_rest():
        raise RESTException(description='error description')

    @app.route('/contenttype', methods=['GET'])
    def test_content_type():
        raise InvalidContentType(allowed_content_types=['application/json'])

    @app.route('/validationerror', methods=['GET'])
    def test_validation_error():
        raise RESTValidationError(
            errors=[FieldError('myfield', 'mymessage', code=10)])

    with app.test_client() as client:
        res = client.get('/')
        assert res.status_code == 200
        data = json.loads(res.get_data(as_text=True))
        assert data['status'] is None
        assert data['message'] == 'error description'

        res = client.get('/contenttype')
        assert res.status_code == 415
        data = json.loads(res.get_data(as_text=True))
        assert data['status'] == 415
        assert 'application/json' in data['message']

        res = client.get('/validationerror')
        assert res.status_code == 400
        data = json.loads(res.get_data(as_text=True))
        print(data)
        assert data['status'] == 400
        assert data['message'] == 'Validation error.'
        assert data['errors'] == [
            dict(field='myfield', message='mymessage', code=10)
        ]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API module for Invenio."""

from __future__ import absolute_import, print_function

from datetime import timezone

from flask import Response, abort, current_app, g, jsonify, make_response, \
    request
from flask.views import MethodView
from werkzeug.exceptions import HTTPException

from .errors import RESTException, SameContentException


def create_api_errorhandler(**kwargs):
    r"""Create an API error handler.

    E.g. register a 404 error:

    .. code-block:: python

        app.errorhandler(404)(create_api_errorhandler(
            status=404, message='Not Found'))

    :param \*\*kwargs: It contains the ``'status'`` and the ``'message'``
        to describe the error.
    """
    def api_errorhandler(e):
        if isinstance(e, RESTException):
            return e.get_response()
        elif isinstance(e, HTTPException) and e.description:
            kwargs['message'] = e.description
        if kwargs.get('status', 400) >= 500 and hasattr(g, 'sentry_event_id'):
            kwargs['error_id'] = str(g.sentry_event_id)
        return make_response(jsonify(kwargs), kwargs['status'])
    return api_errorhandler


class ContentNegotiatedMethodView(MethodView):
    """MethodView with content negotiation.

    Dispatch HTTP requests as MethodView does and build responses using the
    registered serializers. It chooses the right serializer using the request's
    accept type. It also provides a helper method for handling ETags.
    """

    def __init__(self, serializers=None, method_serializers=None,
                 serializers_query_aliases=None, default_media_type=None,
                 default_method_media_type=None, *args, **kwargs):
        """Register the serializing functions.

        Serializing functions will receive all named and non named arguments
        provided to ``make_response`` or returned by request handling methods.
        Recommended prototype is: ``serializer(data, code=200, headers=None)``
        and it should return :class:`flask.Response` instances.

        Serializing functions can also be overridden by setting
        ``self.serializers``.

        :param serializers: A mapping from mediatype to a serializer function.
        :param method_serializers: A mapping of HTTP method name (GET, PUT,
            PATCH, POST, DELETE) -> dict(mediatype -> serializer function). If
            set, it overrides the serializers dict.
        :param serializers_query_aliases: A mapping of values of the defined
            query arg (see `config.REST_MIMETYPE_QUERY_ARG_NAME`) to valid
            mimetypes: dict(alias -> mimetype).
        :param default_media_type: Default media type used if no accept type
            has been provided and global serializers are used for the request.
            Can be None if there is only one global serializer or None. This
            media type is used for method serializers too if
            ``default_method_media_type`` is not set.
        :param default_method_media_type: Default media type used if no accept
            type has been provided and a specific method serializers are used
            for the request. Can be ``None`` if the method has only one
            serializer or ``None``.
        """
        super(ContentNegotiatedMethodView, self).__init__(*args, **kwargs)
        self.serializers = serializers or None
        self.default_media_type = default_media_type
        self.default_method_media_type = default_method_media_type or {}

        # set default default media_types if none has been given
        if self.serializers and not self.default_media_type:
            if len(self.serializers) == 1:
                self.default_media_type = next(iter(self.serializers.keys()))
            elif len(self.serializers) > 1:
                raise ValueError('Multiple serializers with no default media'
                                 ' type')
        # set method serializers
        self.method_serializers = ({key.upper(): func for key, func in
                                    method_serializers.items()} if
                                   method_serializers else {})
        # set serializer aliases
        self.serializers_query_aliases = serializers_query_aliases or {}
        # create default method media_types dict if none has been given
        if self.method_serializers and not self.default_method_media_type:
            self.default_method_media_type = {}
            for http_method, meth_serial in self.method_serializers.items():
                if len(self.method_serializers[http_method]) == 1:
                    self.default_method_media_type[http_method] = \
                        next(iter(self.method_serializers[http_method].keys()))
                elif len(self.method_serializers[http_method]) > 1:
                    # try to use global default media type
                    if default_media_type in \
                            self.method_serializers[http_method]:
                        self.default_method_media_type[http_method] = \
                            default_media_type
                    else:
                        raise ValueError('Multiple serializers for method {0}'
                                         'with no default media type'.format(
                                             http_method))

    def get_method_serializers(self, http_method):
        """Get request method serializers + default media type.

        Grab serializers from ``method_serializers`` if defined, otherwise
        returns the default serializers. Uses GET serializers for HEAD requests
        if no HEAD serializers were specified.

        The method also determines the default media type.

        :param http_method: HTTP method as a string.
        :returns: Tuple of serializers and default media type.
        """
        if http_method == 'HEAD' and 'HEAD' not in self.method_serializers:
            http_method = 'GET'

        return (
            self.method_serializers.get(http_method, self.serializers),
            self.default_method_media_type.get(
                http_method, self.default_media_type)
        )

    def _match_serializers_by_query_arg(self, serializers):
        """Match serializer by query arg."""
        # if the format query argument is present, match the serializer
        arg_name = current_app.config.get('REST_MIMETYPE_QUERY_ARG_NAME')
        if arg_name:
            arg_value = request.args.get(arg_name, None)

            if arg_value is None:
                return None
            # Search for the serializer matching the format
            try:
                return serializers[
                    self.serializers_query_aliases[arg_value]]
            except KeyError:  # either no serializer for this format
                return None

        return None

    def _match_serializers_by_accept_headers(self, serializers,
                                             default_media_type):
        """Match serializer by `Accept` headers."""
        # Bail out fast if no accept headers were given.
        if len(request.accept_mimetypes) == 0:
            return serializers[default_media_type]

        # Determine best match based on quality.
        best_quality = -1
        best = None
        has_wildcard = False
        for client_accept, quality in request.accept_mimetypes:
            if quality <= best_quality:
                continue
            if client_accept == '*/*':
                has_wildcard = True
            for s in serializers:
                if s in ['*/*', client_accept] and quality > 0:
                    best_quality = quality
                    best = s

        # If no match found, but wildcard exists, them use default media
        # type.
        if best is None and has_wildcard:
            best = default_media_type

        if best is not None:
            return serializers[best]
        return None

    def match_serializers(self, serializers, default_media_type):
        """Choose serializer for a given request based on query arg or headers.

        Checks if query arg `format` (by default) is present and tries to match
        the serializer based on the arg value, by resolving the mimetype mapped
        to the arg value.
        Otherwise, chooses the serializer by retrieving the best quality
        `Accept` headers and matching its value (mimetype).

        :param serializers: Dictionary of serializers.
        :param default_media_type: The default media type.
        :returns: Best matching serializer based on `format` query arg first,
            then client `Accept` headers or None if no matching serializer.
        """
        return self._match_serializers_by_query_arg(serializers) or self.\
            _match_serializers_by_accept_headers(serializers,
                                                 default_media_type)

    def make_response(self, *args, **kwargs):
        """Create a Flask Response.

        Dispatch the given arguments to the serializer best matching the
        current request's Accept header.

        :return: The response created by the serializing function.
        :rtype: :class:`flask.Response`
        :raises werkzeug.exceptions.NotAcceptable: If no media type
            matches current Accept header.
        """
        serializer = self.match_serializers(
            *self.get_method_serializers(request.method))

        if serializer:
            return serializer(*args, **kwargs)
        abort(406)

    def dispatch_request(self, *args, **kwargs):
        """Dispatch current request.

        Dispatch the current request using
        :class:`flask.views.MethodView` `dispatch_request()` then, if the
        result is not already a :py:class:`flask.Response`, search for the
        serializing function which matches the best the current request's
        Accept header and use it to build the :py:class:`flask.Response`.

        :rtype: :class:`flask.Response`
        :raises werkzeug.exceptions.NotAcceptable: If no media type matches
            current Accept header.
        :returns: The response returned by the request handler or created by
            the serializing function.
        """
        result = super(ContentNegotiatedMethodView, self).dispatch_request(
            *args, **kwargs
        )

        if isinstance(result, Response):
            return result
        elif isinstance(result, (list, tuple)):
            return self.make_response(*result)
        else:
            return self.make_response(result)

    def check_etag(self, etag, weak=False):
        """Validate the given ETag with current request conditions.

        Compare the given ETag to the ones in the request header If-Match
        and If-None-Match conditions.

        The result is unspecified for requests having If-Match and
        If-None-Match being both set.

        :param str etag: The ETag of the current resource. For PUT and PATCH
            it is the one before any modification of the resource. This ETag
            will be tested with the Accept header conditions. The given ETag
            should not be quoted.
        :raises werkzeug.exceptions.PreconditionFailed: If the
            condition is not met.
        :raises invenio_rest.errors.SameContentException: If the
            the request is GET or HEAD and the If-None-Match condition is not
            met.
        """
        # bool(:py:class:`werkzeug.datastructures.ETags`) is not consistent
        # in Python 3. bool(Etags()) == True even though it is empty.
        if len(request.if_match.as_set(include_weak=weak)) > 0 or \
                request.if_match.star_tag:
            contains_etag = (request.if_match.contains_weak(etag) if weak
                             else request.if_match.contains(etag))
            if not contains_etag and '*' not in request.if_match:
                abort(412)
        if len(request.if_none_match.as_set(include_weak=weak)) > 0 or \
                request.if_none_match.star_tag:
            contains_etag = (request.if_none_match.contains_weak(etag) if weak
                             else request.if_none_match.contains(etag))
            if contains_etag or '*' in request.if_none_match:
                if request.method in ('GET', 'HEAD'):
                    raise SameContentException(etag)
                else:
                    abort(412)

    def check_if_modified_since(self, dt, etag=None):
        """Validate If-Modified-Since with current request conditions."""
        dt = dt.replace(microsecond=0)

        # since Werkzeug v2.0, request-related datetime values are
        # timezone-aware, which compared dates to be timezone-aware as well
        if request.if_modified_since and request.if_modified_since.tzinfo and \
                not dt.tzinfo:
            dt = dt.replace(tzinfo=timezone.utc)

        if request.if_modified_since and dt <= request.if_modified_since:
            raise SameContentException(etag, last_modified=dt)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import json

from flask import Blueprint, Flask, request

from invenio_rest.csrf import REASON_BAD_REFERER, REASON_BAD_TOKEN, \
    REASON_INSECURE_REFERER, REASON_MALFORMED_REFERER, REASON_NO_CSRF_COOKIE, \
    REASON_NO_REFERER, CSRFProtectMiddleware
from invenio_rest.ext import InvenioREST


def test_csrf_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = CSRFProtectMiddleware(app)
    assert 'invenio-csrf' in app.extensions

    app = Flask('testapp')
    ext = CSRFProtectMiddleware()
    assert 'invenio-csrf' not in app.extensions
    ext.init_app(app)
    assert 'invenio-csrf' in app.extensions


def test_csrf_disabled(csrf_app):
    """Test CSRF disabled."""
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200


def test_csrf_enabled(csrf_app, csrf):
    """Test CSRF enabled."""
    with csrf_app.test_client() as client:
        # call a dummy request to obtain the csrf token in the first time
        client.get('/ping')

        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.json['message'] == REASON_BAD_TOKEN
        assert res.status_code == 400

        CSRF_COOKIE_NAME = csrf_app.config['CSRF_COOKIE_NAME']
        CSRF_HEADER_NAME = csrf_app.config['CSRF_HEADER']

        cookie = next(
            (cookie for cookie in client.cookie_jar
                if cookie.name == CSRF_COOKIE_NAME),
            None
        )
        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
            headers={
                CSRF_HEADER_NAME: cookie.value
            }

        )
        assert res.status_code == 200


def test_csrf_before_csrf_protect(csrf_app, csrf):
    """Test before CSRF protect decorator."""
    assert csrf._before_protect_funcs == []

    @csrf.before_csrf_protect
    def before_protect():
        pass

    assert csrf._before_protect_funcs == [before_protect]

    csrf.before_csrf_protect(before_protect)

    assert csrf._before_protect_funcs == [before_protect, before_protect]


def test_csrf_exempt(csrf_app, csrf):
    """Test before CSRF protect decorator."""

    # Test `exempt` as a function passing the name of the view as string
    csrf.exempt('conftest.csrf_test')
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200

    # Test `exempt` as a decorator on a view
    @csrf_app.route('/another-csrf-protect', methods=['POST'])
    @csrf.exempt
    def another_csrf_test():
        return 'another test'

    with csrf_app.test_client() as client:
        res = client.post(
            '/another-csrf-protect',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200

    # Test `exempt` as a decorator on a blueprint
    blueprint = Blueprint("test_csrf_bp", __name__, url_prefix="")

    @blueprint.route('/csrf-protect-bp', methods=['POST'])
    def csrf_bp():
        return 'csrf bp test'

    @blueprint.route('/csrf-protect-bp-2', methods=['POST'])
    def csrf_bp_2():
        return 'csrf bp test 2'

    csrf_app.register_blueprint(blueprint)

    csrf.exempt(blueprint)

    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protect-bp',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200

        res = client.post(
            '/csrf-protect-bp-2',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200


def test_skip_csrf_check(csrf_app, csrf):
    """Test skipping CSRF check."""
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.json['message'] == REASON_NO_CSRF_COOKIE
        assert res.status_code == 400

        @csrf.before_csrf_protect
        def csrf_skip():
            request.skip_csrf_check = True

        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json'
        )
        assert res.status_code == 200


def test_csrf_not_signed_correctly(csrf_app, csrf):
    """Test CSRF malicious attempt with passing malicious cookie and header."""
    from itsdangerous import TimedJSONWebSignatureSerializer

    from invenio_rest.errors import RESTCSRFError

    with csrf_app.test_client() as client:
        # try to pass our own signed cookie and header in an attempt to bypass
        # the CSRF check
        csrf_serializer = TimedJSONWebSignatureSerializer('my_secret')
        malicious_cookie = csrf_serializer.dumps(
            {'user': 'malicious'}, 'my_secret')
        CSRF_COOKIE_NAME = csrf_app.config['CSRF_COOKIE_NAME']
        client.set_cookie('localhost', CSRF_COOKIE_NAME, malicious_cookie)

        res = client.post(
            '/csrf-protected',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
            headers={
                'X-CSRF-Token': malicious_cookie
            },
        )

        assert res.json['message'] == RESTCSRFError.description
        assert res.status_code == 400


def test_csrf_no_referer(csrf_app, csrf):
    """Test CSRF no referrer in a secure request."""
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            base_url='https://localhost',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
        )
        assert res.json['message'] == REASON_NO_REFERER
        assert res.status_code == 400


def test_csrf_malformed_referer(csrf_app, csrf):
    """Test CSRF malformed referrer in a secure request."""
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            base_url='https://localhost',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
            headers={
                'Referer': 'bad-referer'
            }
        )
        assert res.json['message'] == REASON_MALFORMED_REFERER
        assert res.status_code == 400


def test_csrf_insecure_referer(csrf_app, csrf):
    """Test CSRF insecure referrer in a secure request."""
    with csrf_app.test_client() as client:
        res = client.post(
            '/csrf-protected',
            base_url='https://localhost',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
            headers={
                'Referer': 'http://insecure-referer'
            }
        )
        assert res.json['message'] == REASON_INSECURE_REFERER
        assert res.status_code == 400


def test_csrf_bad_referer(csrf_app, csrf):
    """Test CSRF bad referrer in a secure request."""
    with csrf_app.test_client() as client:
        csrf_app.config['APP_ALLOWED_HOSTS'] = [
            'allowed-referer'
        ]
        not_allowed_referer = 'https://not-allowed-referer'
        res = client.post(
            '/csrf-protected',
            base_url='https://localhost',
            data=json.dumps(dict(foo='bar')),
            content_type='application/json',
            headers={
                'Referer': not_allowed_referer
            }
        )
        assert res.json['message'] == REASON_BAD_REFERER % not_allowed_referer
        assert res.status_code == 400



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Exceptions used in Invenio REST module."""

from __future__ import absolute_import, print_function

import json

from flask import g
from werkzeug.exceptions import HTTPException
from werkzeug.http import http_date


class FieldError(object):
    """Represents a field level error.

    .. note:: This is not an actual exception.
    """

    def __init__(self, field, message, code=None):
        """Init object.

        :param field: Field name.
        :param message: The text message to show.
        :param code: The HTTP status to return. (Default: ``None``)
        """
        self.res = dict(field=field, message=message)
        if code:
            self.res['code'] = code

    def to_dict(self):
        """Convert to dictionary.

        :returns: A dictionary with field, message and, if initialized, the
            HTTP status code.
        """
        return self.res


class RESTException(HTTPException):
    """HTTP Exception delivering JSON error responses."""

    errors = None

    def __init__(self, errors=None, **kwargs):
        """Initialize RESTException."""
        super(RESTException, self).__init__(**kwargs)
        if errors is not None:
            self.errors = errors

    def get_errors(self):
        """Get errors.

        :returns: A list containing a dictionary representing the errors.
        """
        return [e.to_dict() for e in self.errors] if self.errors else None

    def get_description(self, environ=None, scope=None):
        """Get the description."""
        return self.description

    def get_body(self, environ=None, scope=None):
        """Get the request body."""
        body = dict(
            status=self.code,
            message=self.get_description(environ),
        )

        errors = self.get_errors()
        if self.errors:
            body['errors'] = errors

        if self.code and (self.code >= 500) and hasattr(g, 'sentry_event_id'):
            body['error_id'] = str(g.sentry_event_id)

        return json.dumps(body)

    def get_headers(self, environ=None, scope=None):
        """Get a list of headers."""
        return [('Content-Type', 'application/json')]


class InvalidContentType(RESTException):
    """Error for when an invalid Content-Type is provided."""

    code = 415
    """HTTP Status code."""

    def __init__(self, allowed_content_types=None, **kwargs):
        """Initialize exception."""
        super(InvalidContentType, self).__init__(**kwargs)
        self.allowed_content_types = allowed_content_types
        self.description = \
            "Invalid 'Content-Type' header. Expected one of: {0}".format(
                ", ".join(allowed_content_types))


class RESTValidationError(RESTException):
    """A standard REST validation error."""

    code = 400
    """HTTP Status code."""

    description = 'Validation error.'
    """Error description."""


class RESTCSRFError(RESTException):
    """A standard REST validation error."""

    code = 400
    """HTTP Status code."""

    description = 'CSRF error.'
    """Error description."""


class SameContentException(RESTException):
    """304 Same Content exception.

    Exception thrown when request is GET or HEAD, ETag is If-None-Match and
    one or more of the ETag values match.
    """

    code = 304
    """HTTP Status code."""

    description = 'Same Content.'
    """Error description."""

    def __init__(self, etag, last_modified=None, **kwargs):
        """Constructor.

        :param etag: matching etag
        :param last_modified: The last modefied date. (Default: ``None``)
        """
        super(SameContentException, self).__init__(**kwargs)
        self.etag = etag
        self.last_modified = last_modified

    def get_response(self, environ=None, scope=None):
        """Get a list of headers."""
        response = super(SameContentException, self).get_response(
            environ=environ
        )
        response.cache_control.no_cache = True
        if self.etag is not None:
            response.set_etag(self.etag)
        if self.last_modified is not None:
            response.headers['Last-Modified'] = http_date(self.last_modified)
        return response



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API module for Invenio.

Invenio-REST takes care of installing basic error handling on a Flask API
application, as well as initializing Flask-Limiter for rate limiting and
Flask-CORS for Cross-Origin Resources Sharing (not enabled by default).
It also acts as orchestrator to take actions depending on the request headers,
such as dealing with ETag/If-Modified-Since cache header or Content-Type/Accept
header to resolve the right content serializer.

Initialization
--------------
First create a Flask application:

    >>> from flask import Flask
    >>> app = Flask('myapp')

Next, initialize your extension:

    >>> from invenio_rest import InvenioREST
    >>> InvenioREST(app)
    <invenio_rest.ext.InvenioREST ...>

Serializers
-----------
Let's create 2 serializers that will return our answer to the correct format.
For instance, our server will be able to either answer in JSON or in XML:

    >>> import xmltodict
    >>> from flask import jsonify, make_response
    >>> def json_v1_search(search_result):
    ...     return make_response(jsonify(search_result))
    >>> def xml_v1_search(search_result):
    ...     return make_response(xmltodict.unparse((search_result,)))

Views
-----
Now we create our view that will handle the requests and return the serialized
response based on the request's headers or using the default media type.
To do so, we need to create a class that inherits
:class:`~.views.ContentNegotiatedMethodView`. In the constructor, we register
our two serializers, and we create a `get` method for the `GET` requests:

    >>> from invenio_rest import ContentNegotiatedMethodView
    >>> class RecordsListResource(ContentNegotiatedMethodView):
    ...     def __init__(self, **kwargs):
    ...         super(RecordsListResource, self).__init__(
    ...             method_serializers={
    ...                 'GET': {
    ...                     'application/json': json_v1_search,
    ...                     'application/xml': xml_v1_search,
    ...                 },
    ...             },
    ...             default_method_media_type={
    ...                 'GET': 'application/json',
    ...             },
    ...             default_media_type='application/json',
    ...             **kwargs)
    ...     def get(self, **kwargs):
    ...         return {"title": "Test"}

To finish, we need to create a blueprint that defines an endpoint (here
`/records`) and that registers our class:

    >>> from flask import Blueprint
    >>> blueprint = Blueprint(
    ...     'mymodule',
    ...     'myapp',
    ...     url_prefix='/records',
    ...     template_folder='templates',
    ...     static_folder='static',
    ... )
    >>> records_view = RecordsListResource.as_view('records')
    >>> blueprint.add_url_rule('/', view_func=records_view)
    >>> app.register_blueprint(blueprint)

Now you can launch your server and request it on the `/records` endpoint

Building REST APIs for Invenio
------------------------------
Following is a quick overview over which tools we are currently using for
building REST APIs. Before implementing your REST API, do take a look at some
of the existing REST APIs already implemented in Invenio to get some
inspiration.

In Invenio we have decided not to use some of the existing Flask extensions for
building REST APIs since mostly these extensions are not very flexible and
there are many existing Python libraries that do a much better job at the
individual tasks.

Flask application
~~~~~~~~~~~~~~~~~
Invenio's REST API is running in its own Flask application. This ensures that
the REST API can run on machines independently of the UI application and also
ensures that e.g. error handling, that it can be independently versioned, is
much simpler compared to having a mixed REST API/UI application.

Views
~~~~~
Views for REST APIs are built using standard Flask blueprints. We use
:class:`~flask.views.MethodView` for HTTP method based dispatching, and in
particular we use Invenio-REST's subclass
:class:`~.views.ContentNegotiatedMethodView` which takes care of selecting the
right serializer based on HTTP content negotiation.

Versioning
~~~~~~~~~~
Versioning of the REST API is primarily achieved through HTTP content
negotiation. I.e. you define a new MIME type that your clients explicitly
request (using :class:`~.views.ContentNegotiatedMethodView`).

Serialization/Deserialization
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Invenio-REST provides 2 ways to define how to serialize the response content:

    - `Query argument`: if the request query contains a well configured
      parameter, for example `format` (see the config
      `REST_MIMETYPE_QUERY_ARG_NAME`), then the serializer associated to the
      value of that argument will be used. With a request like::

          /api/record/<id>?format=custom_json

      and a defined mapping like::

          record_serializers_aliases={
              'custom_json': 'application/json',
              'marc21': 'application/marcxml+xml'
          }

      the output will be serialized using the serializer that is associated
      to the mimetype `application/json`.

    - `Headers`: if the query argument is missing or disabled, then the headers
      `Accept` or `Content-Type` are parsed to resolve the serializer to use.
      The expected value for the header is the mimetype::

          Accept: application/json

      again, the serializer associated to that mimetype will be used to format
      the output.


For serialization/deserialization we primarily use
`Marshmallow <http://marshmallow.readthedocs.io/>`_ which takes care
that REST API endpoints are supplied with proper argument types such as list of
strings or integers, or ensuring that e.g. timestamps are ISO8601 formatted in
UTC when serializing to JSON, and correctly deserialize timestamps into Python
datetime objects.

`Invenio-Records-REST <http://invenio-records-rest.readthedocs.io/>`_
is currently the most advanced example of using serializers with both JSON, XML
and text output.

Request parameters parsing
~~~~~~~~~~~~~~~~~~~~~~~~~~
Request parameters in the URL or JSON are most often handled with the library
`webargs <https://webargs.readthedocs.io/>`_.

Error handling
~~~~~~~~~~~~~~
Invenio-REST provides some default exceptions which you can subclass, which
when thrown, will render a proper REST API response for the error to the
client (see e.g. :class:`~invenio_rest.errors.RESTException`).

Headers (security, CORS and rate limiting)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
`Invenio-App <https://invenio-app.readthedocs.io/>`_ is responsible
for installing
`Flask-Tailsman <https://github.com/GoogleCloudPlatform/flask-talisman>`_ which
sets many important security related headers as well as
`Flask-Limiter <https://flask-limiter.readthedocs.io/>`_ which
provides rate limiting.

Invenio-REST is responsible for installing
`Flask-CORS <https://flask-cors.readthedocs.io/>`_ which provides
support for Cross-Origin Resource Sharing.

`Invenio-OAuth2Server
<https://invenio-oauth2server.readthedocs.io/en/latest/>`_ along with
`Invenio-Accounts <https://invenio-accounts.readthedocs.io/en/latest/>`_ is
responsible for providing API authentication based on OAuth 2.0 as well as
protecting against CRSF-attacks in the REST API.
"""

from __future__ import absolute_import, print_function

from .csrf import csrf
from .ext import InvenioREST
from .version import __version__
from .views import ContentNegotiatedMethodView

__all__ = ('__version__', 'csrf', 'InvenioREST', 'ContentNegotiatedMethodView')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test serializer."""

from __future__ import absolute_import, print_function

from collections import namedtuple

from marshmallow import fields

from invenio_rest.serializer import BaseSchema, result_wrapper


def test_serialize_pretty(app):
    """Test pretty JSON."""
    class TestSchema(BaseSchema):
        title = fields.Str(attribute='title')

    data = {'title': 'test'}
    assert (TestSchema().dump(data)).data == {'title': 'test'}


def test_marshmallow_compatibility():
    """Test wrapper class for marshmallow schema compatibility."""
    dict_result = {'test': 1}
    list_result = [{'test': 1}, {'test': 2}]
    old_marshal = namedtuple('MarshalResult', ['data', 'errors'])
    tuple_result = old_marshal({'test': 1}, [{'field': 5}])

    wrapped = result_wrapper(dict_result)

    assert wrapped == dict_result
    assert wrapped.data == dict_result

    wrapped = result_wrapper(list_result)
    assert wrapped == list_result
    assert wrapped.data == list_result

    wrapped = result_wrapper(tuple_result)
    assert wrapped == tuple_result
    assert isinstance(wrapped, tuple)
    assert tuple_result.data == dict_result



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API module for Invenio."""

from __future__ import absolute_import, print_function

import warnings

import pkg_resources

from . import config
from .views import create_api_errorhandler


class InvenioREST(object):
    """Invenio-REST extension."""

    def __init__(self, app=None):
        """Extension initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization.

        Initialize the Rate-Limiter, CORS and error handlers.

        :param app: An instance of :class:`flask.Flask`.
        """
        self.init_config(app)

        # Enable CORS support if desired
        if app.config['REST_ENABLE_CORS']:
            from flask_cors import CORS
            CORS(app)
            # CORS can be configured using CORS_* configuration variables.

        # Enable CSRF support if desired
        if app.config['REST_CSRF_ENABLED']:
            from .csrf import csrf
            csrf.init_app(app)
        else:
            warnings.warn(
                "CSRF validation will be enabled by default in the version"
                " 1.3.x",
                category=FutureWarning, stacklevel=2)

        app.errorhandler(400)(create_api_errorhandler(
            status=400, message='Bad Request'))
        app.errorhandler(401)(create_api_errorhandler(
            status=401, message='Unauthorized'))
        app.errorhandler(403)(create_api_errorhandler(
            status=403, message='Forbidden'))
        app.errorhandler(404)(create_api_errorhandler(
            status=404, message='Not Found'))
        app.errorhandler(405)(create_api_errorhandler(
            status=405, message='Method Not Allowed'))
        app.errorhandler(406)(create_api_errorhandler(
            status=406, message='Not Acceptable'))
        app.errorhandler(409)(create_api_errorhandler(
            status=409, message='Conflict'))
        app.errorhandler(410)(create_api_errorhandler(
            status=410, message='Gone'))
        app.errorhandler(412)(create_api_errorhandler(
            status=412, message='Precondition Failed'))
        app.errorhandler(415)(create_api_errorhandler(
            status=415, message='Unsupported media type'))
        app.errorhandler(422)(create_api_errorhandler(
            status=422, message='Unprocessable Entity'))
        app.errorhandler(429)(create_api_errorhandler(
            status=429, message='Rate limit exceeded'))
        app.errorhandler(500)(create_api_errorhandler(
            status=500, message='Internal Server Error'))
        app.errorhandler(501)(create_api_errorhandler(
            status=501, message='Not Implemented'))
        app.errorhandler(502)(create_api_errorhandler(
            status=502, message='Bad Gateway'))
        app.errorhandler(503)(create_api_errorhandler(
            status=503, message='Service Unavailable'))
        app.errorhandler(504)(create_api_errorhandler(
            status=504, message='Gateway Timeout'))

        app.extensions['invenio-rest'] = self

    def init_config(self, app):
        """Initialize configuration.

        .. note:: Change Flask-CORS and Flask-Limiter defaults.

        :param app: An instance of :class:`flask.Flask`.
        """
        config_apps = ['REST_', 'CORS_', ]
        for k in dir(config):
            if any([k.startswith(prefix) for prefix in config_apps]):
                app.config.setdefault(k, getattr(config, k))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CSRF Middleware.

The implementation is highly inspred from Django's initial implementation
about CSRF protection. For more information you can see here:
<https://github.com/django/django/blob/master/django/middleware/csrf.py>
"""

import re
import secrets
import string

from flask import Blueprint, abort, current_app, request
from itsdangerous import BadSignature, TimedJSONWebSignatureSerializer
from six import string_types
from six.moves.urllib.parse import urlparse

from .errors import RESTCSRFError

REASON_NO_REFERER = "Referer checking failed - no Referer."
REASON_BAD_REFERER = (
    "Referer checking failed - %s does not match any trusted origins."
)
REASON_NO_CSRF_COOKIE = "CSRF cookie not set."
REASON_BAD_TOKEN = "CSRF token missing or incorrect."
REASON_MALFORMED_REFERER = "Referer checking failed - Referer is malformed."
REASON_INSECURE_REFERER = (
    "Referer checking failed - Referer is insecure while host is secure."
)


def _get_csrf_serializer():
    """Note that this serializer is used to encode/decode the CSRF cookie.

    In case you change this implementation bear in mind that the token
    generated must be signed so as to avoid any client-side tampering.
    """
    return TimedJSONWebSignatureSerializer(
        current_app.config.get(
            'CSRF_SECRET',
            current_app.config.get('SECRET_KEY') or 'CHANGE_ME'),
        salt=current_app.config['CSRF_SECRET_SALT'])


def _get_random_string(length, allowed_chars):
    return ''.join(secrets.choice(allowed_chars) for i in range(length))


def _get_new_csrf_token():
    csrf_serializer = _get_csrf_serializer()
    encoded_token = csrf_serializer.dumps(
        _get_random_string(
            current_app.config['CSRF_TOKEN_LENGTH'],
            current_app.config['CSRF_ALLOWED_CHARS'],
        )
    )
    return encoded_token


def _get_csrf_token():
    try:
        csrf_cookie = request.cookies[
            current_app.config['CSRF_COOKIE_NAME']]
    except KeyError:
        return None
    return _decode_csrf(csrf_cookie)


def _decode_csrf(data):
    csrf_serializer = _get_csrf_serializer()
    try:
        return csrf_serializer.loads(data)
    except BadSignature:
        raise RESTCSRFError()


def _set_token(response):
    response.set_cookie(
        current_app.config['CSRF_COOKIE_NAME'],
        _get_new_csrf_token(),
        max_age=current_app.config.get(
            'CSRF_COOKIE_MAX_AGE', 60*60*24*7*52),
        domain=current_app.config.get(
            'CSRF_COOKIE_DOMAIN',
            current_app.session_interface.get_cookie_domain(
                current_app)),
        path=current_app.session_interface.get_cookie_path(
                current_app),
        secure=current_app.config.get('SESSION_COOKIE_SECURE', True),
        httponly=False,
        samesite=current_app.config['CSRF_COOKIE_SAMESITE'],
    )


def _get_submitted_csrf_token():
    header_name = current_app.config['CSRF_HEADER']
    csrf_token = request.headers.get(header_name)
    if csrf_token:
        return csrf_token
    return None


def _is_referer_secure(referer):
    return 'https' in referer.scheme or \
        not current_app.config['CSRF_FORCE_SECURE_REFERER']


def _abort400(reason):
    abort(400, reason)


def csrf_validate():
    """Check CSRF cookie against request headers."""
    if request.is_secure:
        referer = request.referrer

        if referer is None:
            return _abort400(REASON_NO_REFERER)

        referer = urlparse(referer)
        # Make sure we have a valid URL for Referer.
        if '' in (referer.scheme, referer.netloc):
            return _abort400(REASON_MALFORMED_REFERER)

        # Ensure that our Referer is also secure.
        if not _is_referer_secure(referer):
            return _abort400(REASON_INSECURE_REFERER)

        is_hostname_allowed = referer.hostname in \
            current_app.config.get('APP_ALLOWED_HOSTS')
        if not is_hostname_allowed:
            reason = REASON_BAD_REFERER % referer.geturl()
            return _abort400(reason)

    csrf_token = _get_csrf_token()
    if csrf_token is None:
        return _abort400(REASON_NO_CSRF_COOKIE)

    request_csrf_token = _get_submitted_csrf_token()
    if not request_csrf_token:
        _abort400(REASON_BAD_TOKEN)

    decoded_request_csrf_token = _decode_csrf(request_csrf_token)

    if csrf_token != decoded_request_csrf_token:
        return _abort400(REASON_BAD_TOKEN)


def reset_token():
    """Change the CSRF token in use for a request."""
    request.csrf_cookie_needs_reset = True


class CSRFTokenMiddleware():
    """CSRF Token Middleware."""

    def __init__(self, app=None):
        """Middleware initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize middleware extension.

        :param app: An instance of :class:`flask.Flask`.
        """
        app.config.setdefault('CSRF_COOKIE_NAME', 'csrftoken')
        app.config.setdefault('CSRF_HEADER', 'HTTP_X_CSRFTOKEN')
        app.config.setdefault(
            'CSRF_METHODS', ['POST', 'PUT', 'PATCH', 'DELETE'])
        app.config.setdefault('CSRF_TOKEN_LENGTH', 32)
        app.config.setdefault(
            'CSRF_ALLOWED_CHARS', string.ascii_letters + string.digits)
        app.config.setdefault('CSRF_SECRET_SALT', 'invenio-csrf-token')
        app.config.setdefault('CSRF_FORCE_SECURE_REFERER', True)
        app.config.setdefault(
            'CSRF_COOKIE_SAMESITE',
            app.config.get('SESSION_COOKIE_SAMESITE') or 'Lax')

        @app.after_request
        def csrf_send(response):
            is_method_vulnerable = request.method in app.config['CSRF_METHODS']
            cookie_needs_reset = getattr(
                request, 'csrf_cookie_needs_reset', False)
            cookie_is_missing = current_app.config['CSRF_COOKIE_NAME'] not in \
                request.cookies
            if is_method_vulnerable or cookie_needs_reset or cookie_is_missing:
                _set_token(response)
            return response

        app.extensions['invenio-csrf'] = self


class CSRFProtectMiddleware(CSRFTokenMiddleware):
    """CSRF Middleware."""

    def __init__(self, app=None):
        """Middleware initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        self._exempt_views = set()
        self._exempt_blueprints = set()

        self._before_protect_funcs = []
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Initialize middleware extension.

        :param app: An instance of :class:`flask.Flask`.
        """
        super(CSRFProtectMiddleware, self).init_app(app)

        @app.before_request
        def csrf_protect():
            """CSRF protect method."""
            for func in self._before_protect_funcs:
                func()

            is_method_vulnerable = request.method in app.config['CSRF_METHODS']
            if not is_method_vulnerable:
                return

            if request.blueprint in self._exempt_blueprints:
                return

            if hasattr(request, 'skip_csrf_check'):
                return

            view = app.view_functions.get(request.endpoint)
            if view:
                dest = '{0}.{1}'.format(view.__module__, view.__name__)
                if dest in self._exempt_views:
                    return

            return csrf_validate()

    def before_csrf_protect(self, f):
        """Register functions to be invoked before checking csrf.

        The function accepts nothing as parameters.
        """
        self._before_protect_funcs.append(f)
        return f

    def exempt(self, view):
        """Mark a view or blueprint to be excluded from CSRF protection."""
        if isinstance(view, Blueprint):
            self._exempt_blueprints.add(view.name)
            return view

        if isinstance(view, string_types):
            view_location = view
        else:
            view_location = '.'.join((view.__module__, view.__name__))

        self._exempt_views.add(view_location)
        return view


csrf = CSRFProtectMiddleware()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio REST configuration.

Please also see
`Flask-CORS <https://flask-cors.readthedocs.io/en/latest/>`__ for many more
configuration options.
"""

from __future__ import unicode_literals

CORS_RESOURCES = '*'
"""Dictionary for configuring CORS for endpoints.

   See Flask-CORS for further details.

.. note:: Overwrites
   `Flask-CORS
   <https://flask-cors.readthedocs.io/en/latest/api.html#flask_cors.CORS>`_
   configuration.
"""

CORS_SEND_WILDCARD = True
"""Sending wildcard CORS header.

.. note:: Overwrites
   `Flask-CORS
   <https://flask-cors.readthedocs.io/en/latest/api.html#flask_cors.CORS>`_
   configuration.
"""

CORS_EXPOSE_HEADERS = [
    'ETag',
    'Link',
    'X-RateLimit-Limit',
    'X-RateLimit-Remaining',
    'X-RateLimit-Reset',
    'Content-Type',
]
"""Expose the following headers.

.. note:: Overwrites
   `Flask-CORS
   <https://flask-cors.readthedocs.io/en/latest/api.html#flask_cors.CORS>`_
   configuration.
"""

REST_ENABLE_CORS = False
"""Enable CORS configuration. (Default: ``False``)."""

REST_MIMETYPE_QUERY_ARG_NAME = None
"""Name of the query argument to specify the mimetype wanted for the output.
   Set it to None to disable.

.. note::
   You can customize the query argument name by specifying it as a string::

        REST_MIMETYPE_QUERY_ARG_NAME = 'format'

   With this value, the url will be::

        /api/record/<id>?format=<value>

   You can set the accepted values passing a dictionary to the key
   `record_serializers_aliases`::

       record_serializers_aliases={
          'json': 'application/json',
          'marc21': 'application/marcxml+xml'
       }
"""

REST_CSRF_ENABLED = False
"""Enable CSRF middleware. (Default: ``False``).

.. note::
   The CSRF middleware accepts some configuration parameters that are used to
   adjust the workflow of the CSRF validation. The available options are:

   \\`CSRF_METHODS\\`: HTTP methods against which the csrf check should run.
   Defaults to \\`['POST', 'PUT', 'PATCH', 'DELETE']\\`.

   \\`CSRF_HEADER\\`: The name of the request header used for CSRF
   authentication. Defaults to \\`X-CSRF-Token\\`.

   \\`CSRF_COOKIE_NAME\\`: The name of the request cookie used for CSRF
   authentication. Defaults to \\`_csrftoken\\`.

   \\`CSRF_COOKIE_MAX_AGE\\`: The maximum time until the cookie expires. After
   the expiration the cookie will be removed. Defaults to \\`60*60*24*7*52\\`
   (1 year).

   \\`CSRF_COOKIE_DOMAIN\\`: The domain for which the CSRF cookie should be
   valid. Defaults to \\`flask.sessions.SessionInterface.get_cookie_domain\\`.

   \\`CSRF_COOKIE_PATH\\`: The url path for which the cookie is set. This is
   useful if you have multiple Flask instances running under the same hostname.
   They can use different cookie paths, and each instance will only see its own
   CSRF cookie.
   Defaults to \\`flask.sessions.SessionInterface.get_cookie_path\\`.

   \\`CSRF_COOKIE_SAMESITE\\`: Restrict if CSRF cookie should be sent along
   requests coming from external sites. Defaults to
   \\`SESSION_COOKIE_SAMESITE\\` configuation variable, if this is set to
   a not \\`None\\` value, or \\`Lax\\`. Lax prevents sending cookies with
   CSRF-prone requests from external sites, such as submitting a form.

   \\`CSRF_SECRET\\`: Secret key to encode/decode csrf token. Defaults to
   application \\`SECRET_KEY\\`.

   \\`CSRF_SECRET_SALT\\`: The salt value used to encode/decode csrf token.
   Defaults to \\`invenio-csrf-token\\`.

   \\`CSRF_TOKEN_LENGTH\\`: The length of the generated csrf token. Defaults to
   \\`12\\`.

   \\`CSRF_ALLOWED_CHARS\\`: The allowed characters that can be included in the
   generation of the csrf token. Defaults to \\`string.ascii_letters\\` +
   \\`string.digits\\`.

   \\`CSRF_FORCE_SECURE_REFERER\\`: Flag to disable secure referrer check. This
   should used only in development if you run your UI application over HTTP.
   Defaults to \\`True\\`.
"""
